@@ -14,6 +14,7 @@ use std::{
     ffi::CStr,
     fmt::{self, Debug, Display, Formatter},
     mem::MaybeUninit,
+    str::Utf8Error,
 };
 
 /// Raw [`libxm`] functions.
@@ -94,12 +95,13 @@ impl Context {
 
     /// Play the module and put a generated samples in the `samples` slice.
     ///
-    /// Requires mutability because generating samples moves a playback cursor within context.
+    /// Creates silence if the loop count specified by [`set_max_loop_count`] is reached. By default plays endlessly.
     ///
-    /// Create silence if the end of the module or the loop count specified by [`set_max_loop_count`] is reached.
+    /// Requires mutability because generating samples moves a playback cursor within context.
     ///
     /// [`set_max_loop_count`]: Self::set_max_loop_count
     ///
+    /// # Example
     /// See [`examples`] repository folder for comprehensive examples.
     ///
     /// [`examples`]: https://gitlab.com/fluidex/exmod/-/tree/master/examples
@@ -120,6 +122,10 @@ impl Context {
     }
 
     /// Set the max number of playback loops of the module.
+    ///
+    /// Set to 0 to loop indefinitely.
+    /// By default max loop count is 0 (i. e. infinity).
+    ///
     /// See also [`current_loop_number`].
     ///
     /// [`current_loop_number`]: Self::current_loop_number
@@ -156,8 +162,9 @@ impl Context {
         unsafe { xm_get_number_of_instruments(self.0) }
     }
 
-    /// Get the number of samples of specific instrument.
-    pub fn number_of_samples(&self, instrument: u16) -> Result<u16, OutOfBoundsError> {
+    /// Get the number of samples of the given instrument.
+    pub fn number_of_samples(&self, mut instrument: u16) -> Result<u16, OutOfBoundsError> {
+        instrument += 1;
         if instrument > self.number_of_instruments() {
             Err(OutOfBoundsError::Instruments)
         } else {
@@ -165,8 +172,11 @@ impl Context {
         }
     }
 
-    /// Get the panning of specific channel.
-    pub fn panning_of_channel(&self, channel: u16) -> Result<f32, OutOfBoundsError> {
+    /// Get the panning of the given channel.
+    ///
+    /// Returns a panning between 0 (L) and 1 (R).
+    pub fn panning_of_channel(&self, mut channel: u16) -> Result<f32, OutOfBoundsError> {
+        channel += 1;
         if channel > self.number_of_channels() {
             Err(OutOfBoundsError::Channels)
         } else {
@@ -174,8 +184,11 @@ impl Context {
         }
     }
 
-    /// Get the volume of specific channel.
-    pub fn volume_of_channel(&self, channel: u16) -> Result<f32, OutOfBoundsError> {
+    /// Get the volume of the given channel.
+    ///
+    /// Returns a volume between 0 or 1.
+    pub fn volume_of_channel(&self, mut channel: u16) -> Result<f32, OutOfBoundsError> {
+        channel += 1;
         if channel > self.number_of_channels() {
             Err(OutOfBoundsError::Channels)
         } else {
@@ -183,17 +196,23 @@ impl Context {
         }
     }
 
-    /// Get the frequency of specific channel.
-    pub fn frequency_of_channel(&self, channel: u16) -> Result<f32, OutOfBoundsError> {
-        if channel > self.number_of_channels() {
-            Err(OutOfBoundsError::Channels)
-        } else {
-            unsafe { Ok(xm_get_frequency_of_channel(self.0, channel)) }
-        }
+    /// Get the frequency in Hz of the given channel.
+    pub fn frequency_of_channel(&self, mut channel: u16) -> Result<f32, ChannelError> {
+        channel += 1;
+        self.is_channel_active(channel)
+            .map_err(ChannelError::OutOfBounds)
+            .and_then(|active| {
+                if active {
+                    Ok(unsafe { xm_get_frequency_of_channel(self.0, channel) })
+                } else {
+                    Err(ChannelError::NotActive)
+                }
+            })
     }
 
     /// Check if the channel is active.
-    pub fn is_channel_active(&self, channel: u16) -> Result<bool, OutOfBoundsError> {
+    pub fn is_channel_active(&self, mut channel: u16) -> Result<bool, OutOfBoundsError> {
+        channel += 1;
         if channel > self.number_of_channels() {
             Err(OutOfBoundsError::Channels)
         } else {
@@ -201,8 +220,15 @@ impl Context {
         }
     }
 
-    /// Get the latest trigger of specific instrument.
-    pub fn latest_trigger_of_instrument(&self, instrument: u16) -> Result<u64, OutOfBoundsError> {
+    /// Get the latest trigger of the given instrument.
+    ///
+    /// Get the latest time (in number of generated samples) when
+    /// the given instrument was triggered in any channel.
+    pub fn latest_trigger_of_instrument(
+        &self,
+        mut instrument: u16,
+    ) -> Result<u64, OutOfBoundsError> {
+        instrument += 1;
         if instrument > self.number_of_instruments() {
             Err(OutOfBoundsError::Instruments)
         } else {
@@ -210,12 +236,16 @@ impl Context {
         }
     }
 
-    /// Get the latest trigger of specific sample of specific instrument.
+    /// Get the latest trigger of the given sample of the given instrument.
+    ///
+    /// Get the latest time (in number of generated samples) when
+    /// the given sample of the given instrument was triggered in any channel.
     pub fn latest_trigger_of_sample(
         &self,
-        instrument: u16,
+        mut instrument: u16,
         sample: u16,
     ) -> Result<u64, OutOfBoundsError> {
+        instrument += 1;
         if let Ok(samples) = self.number_of_samples(instrument) {
             if sample > samples {
                 Err(OutOfBoundsError::Samples)
@@ -227,8 +257,12 @@ impl Context {
         }
     }
 
-    /// Get the latest trigger of specific channel.
-    pub fn latest_trigger_of_channel(&self, channel: u16) -> Result<u64, OutOfBoundsError> {
+    /// Get the latest trigger of the given channel.
+    ///
+    /// Get the latest time (in number of generated samples) when
+    /// any instrument was triggered in the given channel.
+    pub fn latest_trigger_of_channel(&self, mut channel: u16) -> Result<u64, OutOfBoundsError> {
+        channel += 1;
         if channel > self.number_of_channels() {
             Err(OutOfBoundsError::Channels)
         } else {
@@ -271,8 +305,9 @@ impl Context {
         }
     }
 
-    /// Mute the specific channel.
-    pub fn mute_channel(&mut self, channel: u16, mute: bool) -> Result<bool, OutOfBoundsError> {
+    /// Mute the given channel.
+    pub fn mute_channel(&mut self, mut channel: u16, mute: bool) -> Result<bool, OutOfBoundsError> {
+        channel += 1;
         if channel > self.number_of_channels() {
             Err(OutOfBoundsError::Channels)
         } else {
@@ -280,12 +315,13 @@ impl Context {
         }
     }
 
-    /// Mute the specific instrument.
+    /// Mute the given instrument.
     pub fn mute_instrument(
         &mut self,
-        instrument: u16,
+        mut instrument: u16,
         mute: bool,
     ) -> Result<bool, OutOfBoundsError> {
+        instrument += 1;
         if instrument > self.number_of_instruments() {
             Err(OutOfBoundsError::Instruments)
         } else {
@@ -305,25 +341,21 @@ impl Context {
 #[cfg(feature = "strings")]
 impl Context {
     /// Get the module name.
-    ///
-    /// # Safety
-    /// The library relies that the module name will be encoded in UTF-8. If not, UB or invalid string is possible.
-    pub unsafe fn module_name(&self) -> String {
-        CStr::from_ptr(xm_get_module_name(self.0))
-            .to_string_lossy()
-            .trim_end_matches(' ')
-            .into()
+    pub fn module_name(&self) -> Result<&str, Utf8Error> {
+        unsafe {
+            Ok(CStr::from_ptr(xm_get_module_name(self.0))
+                .to_str()?
+                .trim_end_matches(' '))
+        }
     }
 
     /// Get the the name of the tracker in which the module was created.
-    ///
-    /// # Safety
-    /// The library relies that the tracker name will be encoded in UTF-8. If not, UB or invalid string is possible.
-    pub unsafe fn tracker_name(&self) -> String {
-        CStr::from_ptr(xm_get_tracker_name(self.0))
-            .to_string_lossy()
-            .trim_end_matches(' ')
-            .into()
+    pub fn tracker_name(&self) -> Result<&str, Utf8Error> {
+        unsafe {
+            Ok(CStr::from_ptr(xm_get_tracker_name(self.0))
+                .to_str()?
+                .trim_end_matches(' '))
+        }
     }
 }
 
@@ -390,3 +422,19 @@ impl Display for OutOfBoundsError {
         })
     }
 }
+
+/// Error for [`Context::frequency_of_channel`] function.
+#[derive(Debug, PartialEq, Eq, Clone, Copy)]
+pub enum ChannelError {
+    OutOfBounds(OutOfBoundsError),
+    NotActive,
+}
+
+impl Display for ChannelError {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::OutOfBounds(error) => Display::fmt(error, f),
+            Self::NotActive => f.write_str("Channel is not active!"),
+        }
+    }
+}
